﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Text;
using System.Threading.Tasks;

namespace PristaniNaMesici
{
    public class Game
    {
        public Game(string address, int maxSpeed = 0)
        {
            this.Address = address;
            this.MaxSpeed = maxSpeed <= 0 ? 0 : maxSpeed;

            this.Http = new HttpClient();
        }

        #region Permanent variables

        public string Address
        {
            get;
        }

        /// <summary>
        /// Maximalni rychlost.
        /// Neni urceno zda na osach nebo (smerovem) vektoru.
        /// </summary>
        public int MaxSpeed
        {
            get;
        }

        public HttpClient Http;

        #endregion

        #region (Static) Lander settings

        public const float GravityY = -7;

        public const float EngineUpPower = 12;
        public const float EngineUpFuel = 1;

        public const float EngineSidePower = 6;
        public const float EngineSideFuel = 0.5f;
        
        public const float SizeFromCenterX = 10;
        public const float SizeFromCenterY = 10;

        /// <summary>
        /// Total module width
        /// </summary>
        public const float SizeTotalX = SizeFromCenterX * 2;
        /// <summary>
        /// Total module height
        /// </summary>
        public const float SizeTotalY = SizeFromCenterY * 2;

        #endregion

        public enum CurrentStateEnum
        {
            InAir,
            Crashed,
            Landed
        }

        #region Variables (last valid)

        /// <summary>
        /// Cislo pristani, nemelo by se menit
        /// </summary>
        public uint ID
        {
            get;
            private set;
        } = 0;

        /// <summary>
        /// Aktualni stav, pokud neni <see cref="CurrentStateEnum.InAir"/> nastava konec simulace
        /// </summary>
        public CurrentStateEnum CurrentState
        {
            get;
            private set;
        } = CurrentStateEnum.InAir;

        public float X
        {
            get;
            private set;
        }
        public float Y
        {
            get;
            private set;
        }

        /// <summary>
        /// Zmena v rychlosti
        /// </summary>
        public float Dx
        {
            get;
            private set;
        }
        /// <summary>
        /// Zmena v rychlosti
        /// </summary>
        public float Dy
        {
            get;
            private set;
        }

        /// <summary>
        /// Mnozstvi paliva
        /// </summary>
        public float Fuel
        {
            get;
            private set;
        } = 0;

        /// <summary>
        /// Leva souradnice pristavaci plochy
        /// </summary>
        public float LandingLeft
        {
            get;
            private set;
        }
        /// <summary>
        /// Prava souradnice pristavaci plochy
        /// </summary>
        public float LandingRight
        {
            get;
            private set;
        }

        /// <summary>
        /// Motor tlacici smerem nahoru
        /// </summary>
        public bool EngineUp
        {
            get;
            private set;
        } = false;
        /// <summary>
        /// Motor tlacici smerem vlevo
        /// </summary>
        public bool EngineLeft
        {
            get;
            private set;
        } = false;
        /// <summary>
        /// Motor tlacici smerem vpravo
        /// </summary>
        public bool EngineRight
        {
            get;
            private set;
        } = false;

        #endregion

        #region Calculated

        /// <summary>
        /// Maximalni hodnota paliva (ktera kdy byla dosazena).
        /// </summary>
        public float MaxFuel
        {
            get;
            private set;
        }

        /// <summary>
        /// Sirka pristavaci plochy.
        /// </summary>
        public float LandingSize
        {
            get
            {
                return LandingRight - LandingLeft;
            }
        }

        /// <summary>
        /// Stredovy bod pristavaci plochy.
        /// </summary>
        public float LandingCenter
        {
            get
            {
                return LandingLeft + LandingSize / 2;
            }
        }

        /// <summary>
        /// Overeni minimalni sirky pristavaci plochy (<see cref="SizeTotalX"/>)
        /// </summary>
        public bool LandingValid
        {
            get
            {
                return LandingSize >= SizeTotalX;
            }
        }

        /// <summary>
        /// Levy okraj pristavaci plochy pro stred pristavaciho modulu
        /// </summary>
        public float LandingLeftCenter
        {
            get
            {
                return LandingLeft + SizeFromCenterX;
            }
        }
        /// <summary>
        /// Pravy okraj pristavaci plochy pro stred pristavaciho modulu
        /// </summary>
        public float LandingRightCenter
        {
            get
            {
                return LandingLeft - SizeFromCenterX;
            }
        }

        /// <summary>
        /// Vzdalenost spodku modulu od povrchu
        /// </summary>
        public float DistanceToGroundYBottom
        {
            get
            {
                var dis = Y - 10;
                return dis <= 0 ? 0 : dis;
            }
        }
        /// <summary>
        /// Vzdalenost stredu modulu od povrchu
        /// </summary>
        public float DistanceToGroundYCenter
        {
            get
            {
                return Y <= 0 ? 0 : Y;
            }
        }

        /// <summary>
        /// May return negative values (consider 1D vector)
        /// </summary>
        public float DistanceToLandingCenterX
        {
            get
            {
                return LandingCenter - X;
            }
        }

        /// <summary>
        /// May return negative values (consider 1D vector)
        /// </summary>
        public float MinimumDistanceToValidLandingX
        {
            get
            {
                /*
                if (IsAboveLanding)
                    return 0;
                */
                if (X < LandingLeftCenter)
                    return LandingLeftCenter - X;
                if (X > LandingRightCenter)
                    return LandingRightCenter - X;
                return 0;
            }
        }

        public bool IsAboveLandingEdges
        {
            get
            {
                return X >= LandingLeft && X <= LandingRight;
            }
        }
        public bool IsAboveLandingCenter
        {
            get
            {
                return X >= LandingLeftCenter && X <= LandingRightCenter;
            }
        }

        public float TimeBeforeLanding
        {
            get
            {
                var dist = DistanceToGroundYBottom;
                // y_new = y_old + dY * t + 0.5 * (-7 + 12 * up) * t^2
                // dY_new = dY_old + (-7 + 12 * up) * t

                // y_1s += dY_1s + 0.5 * (-7 + 12 * up)
                // dY_1s += -7 + 12 * up

                // Engine on
                // y = (-7 + 12 * up) * t + 0.5 * (-7 + 12 * up) * t^2
                // t = sqrt((2y / 5) + 1) - 1

                // Engine off
                // y = (-7) * t + 0.5 * (-7) * t^2
                // t = sqrt(1 - (2y/7)) - 1

                return (float)Math.Sqrt(1 - ((2 * dist) / 7)) - 1;
            }
        }

        public float TimeBeforeLandingWithEngineOn
        {
            get
            {
                var dist = DistanceToGroundYBottom;

                return (float)Math.Sqrt(((2 * dist) / 5) + 1) - 1;
            }
        }
        public float SlowdownFuelRequired
        {
            get
            {
                return TimeBeforeLandingWithEngineOn;
            }
        }
        public bool EnoughFuelForSlowdown
        {
            get
            {
                return Fuel >= SlowdownFuelRequired;
            }
        }

        #endregion

        public async Task<bool> UpdateAsync(bool up, bool right, bool left)
        {
            string url;
            // Priprava request url
            {
                StringBuilder url_b = new StringBuilder();

                url_b.Append("http://");
                url_b.Append(Address);
                url_b.Append("/");
                url_b.Append("stav");
                url_b.Append("?");

                url_b.Append("id=");
                url_b.Append(ID);

                if (up != EngineUp)
                {
                    url_b.Append("&");

                    url_b.Append("nahoru=");
                    url_b.Append(up ? 1 : 0);
                }

                if (left != EngineLeft)
                {
                    url_b.Append("&");

                    url_b.Append("doleva=");
                    url_b.Append(left ? 1 : 0);
                }

                if (right != EngineRight)
                {
                    url_b.Append("&");

                    url_b.Append("doprava=");
                    url_b.Append(right ? 1 : 0);
                }

                url = url_b.ToString();
            }

            //Console.WriteLine("URL: '{0}'", url);

            // Nastaveni predchozi stavu motoru
            {
                EngineUp = up;
                EngineRight = right;
                EngineLeft = left;
            }

            using (var response = await Http.GetAsync(url))
            {
                var response_stream = await response.Content.ReadAsStreamAsync();
                var reader = new StreamReader(response_stream);

                ReadDataFromReader(reader);
            }

            return true;
        }

        public async Task<bool> StartAsync()
        {
            string url;
            // Priprava request url
            {
                StringBuilder url_b = new StringBuilder();

                url_b.Append("http://");
                url_b.Append(Address);
                url_b.Append("/");
                url_b.Append("start");

                url = url_b.ToString();
            }

            //Console.WriteLine("URL: '{0}'", url);

            using (var response = await Http.GetAsync(url))
            {
                var response_stream = await response.Content.ReadAsStreamAsync();
                var reader = new StreamReader(response_stream);

                ReadDataFromReader(reader);
            }

            Started = true;

            return true;
        }

        public bool Started
        {
            get;
            private set;
        } = false;

        public void ReadDataFromReader(StreamReader reader)
        {
            string line;

            // Line 0
            // ID or result
            {
                line = reader.ReadLine();

                //Console.WriteLine("Line: '{0}'", line);

                if (line == "Zdar!")
                {
                    CurrentState = CurrentStateEnum.Landed;
                    return;
                }
                if (line == "Krach!")
                {
                    CurrentState = CurrentStateEnum.Crashed;
                    return;
                }
                ID = uint.Parse(line);
            }

            // Line 1
            // X Y
            {
                line = reader.ReadLine();
                var split = line.Split(' ');
                if (split.Length != 2)
                    throw new IndexOutOfRangeException("spatny pocet souradnic");

                X = float.Parse(split[0]);
                Y = float.Parse(split[1]);
            }

            // Line 2
            // Vx Vy
            // dX dY
            {
                line = reader.ReadLine();
                var split = line.Split(' ');
                if (split.Length != 2)
                    throw new IndexOutOfRangeException("spatny pocet souradnic");

                Dx = float.Parse(split[0]);
                Dy = float.Parse(split[1]);
            }

            // Line 3
            // P
            // Fuel
            {
                line = reader.ReadLine();
                Fuel = float.Parse(line);

                if (Fuel > MaxFuel)
                    MaxFuel = Fuel;
            }

            // Line 4
            // L R
            // LandingLeft LendingRight
            {
                line = reader.ReadLine();
                var split = line.Split(' ');
                if (split.Length != 2)
                    throw new IndexOutOfRangeException("spatny pocet souradnic pristavaciho bodu");

                LandingLeft = float.Parse(split[0]);
                LandingRight = float.Parse(split[1]);
            }
        }

        #region Autopilot

        public bool AutopilotActive
        {
            get;
            set;
        } = false;

        public async Task<bool> UpdateByAutopilot()
        {
            // Regulace rychlostik dolu
            {
                if (Dy > 5)
                    EngineUp = true;
                if (Dy < 4)
                    EngineUp = false;
            }

            // Regulace smeru
            {
                if(IsAboveLandingCenter)
                {
                    if(Math.Abs(Dx) > 0.5f)
                    // Moving too fast to sides
                    {
                        if (Dx > 0)
                        {
                            EngineLeft = true;
                            EngineRight = false;
                        }
                        if (Dx < 0)
                        {
                            EngineLeft = false;
                            EngineRight = true;
                        }
                    }
                    else
                    // Moving to sides is OK
                    {
                        EngineLeft = false;
                        EngineRight = false;
                    }
                }
                else
                {
                    // Setup direction
                    var min_dist = MinimumDistanceToValidLandingX;
                    if (min_dist > 0)
                    {
                        EngineLeft = true;
                        EngineRight = false;
                    }
                    if (min_dist < 0)
                    {
                        EngineLeft = false;
                        EngineRight = true;
                    }

                    var curr_dist = Math.Sqrt(Dx * Dx + Dy * Dy);

                    // Check if moving too fast or OK
                    {
                        if (curr_dist > min_dist * 2)
                        {
                            // Moving TOO fast, toggle direction
                            EngineLeft = !EngineLeft;
                            EngineRight = !EngineLeft;
                        }
                        if(curr_dist <= min_dist)
                        {
                            EngineLeft = false;
                            EngineRight = false;
                        }
                    }
                }
            }

            // Omezeni maximalni rychlosti
            if(MaxSpeed != 0)
            {
                if(Math.Abs(Dx) > MaxSpeed)// Prekroceno na ose X
                {
                    if (Dx > 0)
                    {
                        EngineLeft = true;
                        EngineRight = false;
                    }
                    if (Dx < 0)
                    {
                        EngineLeft = false;
                        EngineRight = true;
                    }
                }

                // Diky omezovaci rychlosti neni potreba resit osu Y
            }

            return await UpdateAsync(EngineUp, EngineRight, EngineLeft);
        }

        #endregion

    }
}